// Ryan Schmidt - rms@unknownroad.com
// Copyright (c) 2005-2010
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt


// AnimationManager.cpp : implementation file
//

#include "stdafx.h"
#include "BlobShop.h"
#include "AnimationManager.h"
#include ".\animationmanager.h"

#include "SceneManager.h"
#include "SetFloatValueDialog.h"
#include <WmlExtVectorUtil.h>

using namespace rmsui;

// CAnimationManager

IMPLEMENT_DYNAMIC(CAnimationManager, CWnd)
CAnimationManager::CAnimationManager()
{
	m_nCurFrame = 0;
	m_nMaxFrame = 1;
}

CAnimationManager::~CAnimationManager()
{
}


//! set current time. Will set current frame to nearest frame
void CAnimationManager::SetCurrentTime( float fTime )
{
	SetParameters( fTime );
	m_nCurFrame = (unsigned int)fTime;
}


//! set current frame and update all parameters
void CAnimationManager::SetCurrentFrame( unsigned int nFrame )
{
	SetParameters( (float)nFrame );
	m_nCurFrame = nFrame;
}

//! set a keyframe at current time
void CAnimationManager::SetKeyFrame()
{
	SetKeyFrame( m_nCurFrame );
	m_vKeyframes.insert( m_nCurFrame );
}

void CAnimationManager::ClearAll()
{
	// erase the list of keyframes
	m_vFloatParameters.clear();
	m_vFrame3fParameters.clear();

	// synchronize the parameter list
	SynchronizeParameterList();
}

void CAnimationManager::ClearCurrentKeyFrame()
{
	ClearKeyFrame( m_nCurFrame );
	m_vKeyframes.erase( m_nCurFrame );
}


void CAnimationManager::SetKeyFrame( unsigned int nKeyFrame )
{
	// synchronize the parameter list
	SynchronizeParameterList();

	// do float params

	std::map<IParameter *, FloatValueSet>::iterator 
		mcur( m_vFloatParameters.begin()  ),  mend( m_vFloatParameters.end() );
	while ( mcur != mend ) {
		IParameter * vParam = mcur->first;
		FloatValueSet & vSet = mcur->second;
		++mcur;

		if ( vParam->GetNoKeyFrame() == true )
			continue;

		FloatValue newV( nKeyFrame, vParam->GetValueFloat() );

		// if this keyframe exists already, remove it
		FloatValueSet::iterator found( vSet.find(newV) );
		if ( found != vSet.end() )
			vSet.erase( found );

		vSet.insert( newV );
	}

#if 0
	// do frame so's

	std::map<SceneObject *, Frame3fValueSet>::iterator 
		f3cur( m_vFrame3fParameters.begin()  ),  f3end( m_vFrame3fParameters.end() );
	while ( f3cur != f3end ) {
		SceneObject * vSO = f3cur->first;
		Frame3fValueSet & vSet = f3cur->second;
		++f3cur;

		//if ( vParam->GetNoKeyFrame() == true )
		//	continue;

		Frame3fValue newV( nKeyFrame, vSO->GetFrame() );

		// if this keyframe exists already, remove it
		Frame3fValueSet::iterator found( vSet.find(newV) );
		if ( found != vSet.end() )
			vSet.erase( found );

		vSet.insert( newV );
	}
#endif
}

void CAnimationManager::SynchronizeParameterList()
{
	unsigned int nObjCount = CSceneManager::Get()->GetSceneObjectManager()->GetSceneObjectCount();
	std::vector<SceneObject *> vObjects;
	vObjects.resize(nObjCount);
	if (! CSceneManager::Get()->GetSceneObjectManager()->GetSceneObjects(vObjects) ) {
		ASSERT(0 && "Why?");
		return;
	}

	// ok, for each object, read parameter list. Look for float params. For each
	// float param, store in the set
	std::set< IParameter * > vCurFloatParams;
	size_t nObjects = vObjects.size();
	for ( unsigned int i = 0; i < nObjects; ++i ) {
		const std::vector<IParameter *> & vParams = vObjects[i]->GetParameterSet().GetParameters();
		size_t nParams = vParams.size();
		for ( unsigned int j = 0; j < nParams; ++j ) {
			if ( vParams[j]->GetNoKeyFrame() || vParams[j]->GetType() != IParameter::FloatParameter )
				continue;
			vCurFloatParams.insert( vParams[j] );
		}
	}

	// process float parameters

	// ok, now check that all the params in m_vFloatParameters still exist
	std::map<IParameter *, FloatValueSet>::iterator 
		mcur( m_vFloatParameters.begin()  ),  mend( m_vFloatParameters.end() );
	while ( mcur != mend ) {
		std::map<IParameter *, FloatValueSet>::iterator tmp = mcur++;
		if ( vCurFloatParams.find((*tmp).first) == vCurFloatParams.end() )
			m_vFloatParameters.erase( tmp );
	}

	// now add new parameters
	std::set< IParameter * >::iterator vcur( vCurFloatParams.begin() ), vend( vCurFloatParams.end() );
	while ( vcur != vend ) {
		if ( m_vFloatParameters.find( *vcur ) == m_vFloatParameters.end() ) {
			FloatValueSet newSet;
			newSet.insert( FloatValue( 0, (*vcur)->GetValueFloat() ) );
			m_vFloatParameters[ *vcur ] = newSet;
		}
		++vcur;
	}


	// make list of frame3f values. Note that this is per-sceneobject, there are no parameters for this yet (fuck!)
	
	// make set of SO's so the next bit is easier
	std::set<SceneObject *> vSOSet;
	for ( unsigned int i = 0; i < nObjects; ++i )
		vSOSet.insert( vObjects[i] );

	// ok, now check that all the scene objects in m_vFrame3fParameters still exist
	{
	std::map<SceneObject *, Frame3fValueSet>::iterator 
		mcur( m_vFrame3fParameters.begin()  ),  mend( m_vFrame3fParameters.end() );
	while ( mcur != mend ) {
		std::map<SceneObject *, Frame3fValueSet>::iterator tmp = mcur++;
		if ( vSOSet.find((*tmp).first) == vSOSet.end() )
			m_vFrame3fParameters.erase( tmp );
	}

	// now add new parameters
	std::set< SceneObject * >::iterator vcur( vSOSet.begin() ), vend( vSOSet.end() );
	while ( vcur != vend ) {
		if ( m_vFrame3fParameters.find( *vcur ) == m_vFrame3fParameters.end() ) {
			Frame3fValueSet newSet;
			newSet.insert( Frame3fValue( 0, (*vcur)->GetFrame() ) );
			m_vFrame3fParameters[ *vcur ] = newSet;
		}
		++vcur;
	}
	}		// extra parens so we don't have to rename all the iterators..

}	

//! remove all keys at a current frame
void CAnimationManager::ClearKeyFrame( unsigned int nKeyFrame )
{
	// synchronize the parameter list
	SynchronizeParameterList();

	if ( nKeyFrame == 0 )		// cannot kill frame 0!
		return;

	// do float params

	std::map<IParameter *, FloatValueSet>::iterator 
		mcur( m_vFloatParameters.begin()  ),  mend( m_vFloatParameters.end() );
	while ( mcur != mend ) {
		FloatValueSet & vSet = mcur->second;
		++mcur;

		FloatValue newV( nKeyFrame, 0 );	
		FloatValueSet::iterator found( vSet.find(newV) );
		if ( found != vSet.end() )
			vSet.erase( found );
	}

	// do frame so's

	std::map<SceneObject  *, Frame3fValueSet>::iterator 
		f3cur( m_vFrame3fParameters.begin()  ),  f3end( m_vFrame3fParameters.end() );
	while ( f3cur != f3end ) {
		Frame3fValueSet & vSet = f3cur->second;
		++f3cur;

		Frame3fValue newV( nKeyFrame, Wml::Frame3f() );	
		Frame3fValueSet::iterator found( vSet.find(newV) );
		if ( found != vSet.end() )
			vSet.erase( found );
	}
}


void CAnimationManager::SetParameters( float fTime )
{
	// synchronize the parameter list
	SynchronizeParameterList();

	// iterate through and set all float parameters
	std::map<IParameter *, FloatValueSet>::iterator 
		mcur( m_vFloatParameters.begin()  ),  mend( m_vFloatParameters.end() );
	while ( mcur != mend ) {
		IParameter * vParam = mcur->first;
		FloatValueSet & vSet = mcur->second;
		++mcur;

		// find previous and next keyframe
		FloatValueSet::iterator iPrev(vSet.end()), iNext(vSet.end());
		FloatValueSet::iterator cur( vSet.begin() ), end( vSet.end() );
		while ( cur != end ) {
			if ( (float)cur->nFrame <= fTime ) {
				iPrev = cur;
				cur++;
				iNext = cur;
			} else
				cur = end;		// must be finished
		}
		if ( iPrev == end )		// already hit last keyframe
			continue;

		else if ( iNext == end )	{	// *on* the last keyframe - just set value
			if ( vParam->GetValueFloat() != (*iPrev).value )
				vParam->SetValueFloat( (*iPrev).value );

		} else {		// must be between keyframes - interpolate!
			float fValue = InterpValueLinear( (*iPrev), (*iNext), fTime );
			if ( vParam->GetValueFloat() != fValue )
				vParam->SetValueFloat( fValue );
		}
	}


	// iterate through and set all frame so's
	std::map<SceneObject *, Frame3fValueSet>::iterator 
		f3cur( m_vFrame3fParameters.begin()  ),  f3end( m_vFrame3fParameters.end() );
	while ( f3cur != f3end ) {
		SceneObject * vSO = f3cur->first;
		Frame3fValueSet & vSet = f3cur->second;
		++f3cur;

		// find previous and next keyframe
		Frame3fValueSet::iterator iPrev(vSet.end()), iNext(vSet.end());
		Frame3fValueSet::iterator cur( vSet.begin() ), end( vSet.end() );
		while ( cur != end ) {
			if ( (float)cur->nFrame <= fTime ) {
				iPrev = cur;
				cur++;
				iNext = cur;
			} else
				cur = end;		// must be finished
		}
		if ( iPrev == end )		// already hit last keyframe
			continue;

		else if ( iNext == end )	{	// *on* the last keyframe - just set value
			//if ( vParam->GetValueFloat() != (*iPrev).value )
			//	vParam->SetValueFloat( (*iPrev).value );
			Wml::Matrix3f vRotate = FindRotation( vSO->GetFrame(), (*iPrev).value );
			vSO->Rotate( vRotate );

		} else {		// must be between keyframes - interpolate!
			//float fValue = InterpValueLinear( (*iPrev), (*iNext), fTime );
			//if ( vParam->GetValueFloat() != fValue )
			//	vParam->SetValueFloat( fValue );

			Wml::Frame3f vFrame = InterpFrameLinear( (*iPrev), (*iNext), fTime );
			Wml::Matrix3f vRotate = FindRotation( vSO->GetFrame(), vFrame );
			vSO->Rotate( vRotate );
		}
	}

	// update tool manager...
	CSceneManager::Get()->ForceToolManagerUpdate();

}



float CAnimationManager::InterpValueLinear( const CAnimationManager::FloatValue & vFirst, 
										    const CAnimationManager::FloatValue & vSecond, 
											float fTime )
{
	int nFrameDelta = vSecond.nFrame - vFirst.nFrame;
	float fInterpTime = fTime - (float)vFirst.nFrame;

	float fAlpha = fInterpTime / (float)nFrameDelta;
	return (1.0f - fAlpha)*vFirst.value + fAlpha*vSecond.value;
}


Wml::Frame3f CAnimationManager::InterpFrameLinear( const Frame3fValue & vFirst, const Frame3fValue & vSecond, float fTime )
{
	int nFrameDelta = vSecond.nFrame - vFirst.nFrame;
	float fInterpTime = fTime - (float)vFirst.nFrame;
	float fAlpha = fInterpTime / (float)nFrameDelta;

	// get quaternions for each frame and slerp between them
	Wml::Quaternionf qFirst( vFirst.value.FrameMatrix() );
	Wml::Quaternionf qSecond( vSecond.value.FrameMatrix() );
	Wml::Quaternionf qInterp = Wml::Quaternionf::Slerp( fAlpha, qFirst, qSecond );

	// now get back the rotation matrix and set frame directly to it
	Wml::Matrix3f matRotate;
	qInterp.ToRotationMatrix(matRotate);
	Wml::Frame3f vInterp(vFirst.value.Origin());
	vInterp.SetRotation( matRotate.Transpose() );
	return vInterp;
}


Wml::Matrix3f CAnimationManager::FindRotation( const Wml::Frame3f & vCurFrame, const Wml::Frame3f & vDestFrame )
{
	// align vCurFrame.Z() with vDestFrame.Z()
	Wml::Matrix3f matAlignZ;
	ComputeAlignAxisMatrix( vCurFrame.Z(), vDestFrame.Z(), matAlignZ );
	Wml::Frame3f vCopy( vCurFrame );
	vCopy.Rotate( matAlignZ );

	// compute rotation angle around vDestFrame.Z()
	Wml::Vector3f vX1( vDestFrame.X() );
	Wml::Vector3f vX2( vCopy.X() );
	Wml::Matrix3f matAlignX;
	ComputeAlignAxisMatrix( vX2, vX1, matAlignX );

	return matAlignX * matAlignZ;
}

BEGIN_MESSAGE_MAP(CAnimationManager, CWnd)
	ON_COMMAND(ID_ANIMATION_SETCURRENTFRAME, OnAnimationSetcurrentframe)
	ON_COMMAND(ID_ANIMATION_SETLENGTH, OnAnimationSetlength)
	ON_COMMAND(ID_ANIMATION_RESETANIMATION, OnAnimationResetanimation)
	ON_COMMAND(ID_ANIMATION_SETKEYFRAME, OnAnimationSetkeyframe)
	ON_COMMAND(ID_ANIMATION_CLEARKEYFRAME, OnAnimationClearkeyframe)
END_MESSAGE_MAP()



// CAnimationManager message handlers




void CAnimationManager::OnAnimationSetcurrentframe()
{
	// show modal dialog for frame
	CSetFloatValueDialog dlg(0, (float)65536, (float)m_nCurFrame, "Current Frame", true, NULL);
	if ( dlg.DoModal() == IDOK ) {
		unsigned int nNewFrame = dlg.GetNewValueInteger();
		if ( nNewFrame > m_nMaxFrame )
			m_nMaxFrame = nNewFrame;
		SetCurrentFrame( nNewFrame );
	}
}

void CAnimationManager::OnAnimationSetlength()
{
	// find last keyframe
	std::set<unsigned int>::iterator cur( m_vKeyframes.begin() );
	unsigned int nMax = 0;
	while ( cur != m_vKeyframes.end() )
		nMax = *cur++;

	// show modal dialog for frame
	CSetFloatValueDialog dlg((float)(nMax+1), 65536, (float)m_nMaxFrame, "Max Frame", true, NULL);
	if ( dlg.DoModal() == IDOK ) {
		m_nMaxFrame = dlg.GetNewValueInteger();
	}
}

void CAnimationManager::OnAnimationResetanimation()
{
	ClearAll();
}



void CAnimationManager::OnAnimationSetkeyframe()
{
	SetKeyFrame();
}

void CAnimationManager::OnAnimationClearkeyframe()
{
	ClearCurrentKeyFrame();
}




