// Ryan Schmidt - rms@unknownroad.com
// Copyright (c) 2005-2010
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt


#pragma once

#include <SceneObject.h>
#include <map>
#include <set>
#include <vector>

// CAnimationManager

class CAnimationManager : public CWnd
{
	DECLARE_DYNAMIC(CAnimationManager)

public:
	CAnimationManager();
	virtual ~CAnimationManager();

	//! set current time. Will set current frame to nearest frame
	void SetCurrentTime( float fTime );

	//! set current frame and update all parameters
	void SetCurrentFrame( unsigned int nFrame );

	//! set a keyframe at current time
	void SetKeyFrame();

	//! remove all keyframes
	void ClearAll();

	//! remove keyframe at current time
	void ClearCurrentKeyFrame();

	//! get current frame
	unsigned int GetCurrentFrame() { return m_nCurFrame; }

	//! get last frame
	unsigned int GetMaxFrame() { return m_nMaxFrame; }

protected:
	DECLARE_MESSAGE_MAP()

	unsigned int m_nMaxFrame;

	unsigned int m_nCurFrame;
	std::set<unsigned int> m_vKeyframes;

	template<typename T>
	struct KeyFrameValueType {
		unsigned int nFrame;
		T value;

		KeyFrameValueType()
			{ nFrame = 0; value = T(); }
		KeyFrameValueType( unsigned int frame, const T & initialValue )
			{ nFrame = frame; value = initialValue; }
		bool operator<( const KeyFrameValueType<T> & v2 ) const {
			return nFrame <  v2.nFrame;
		}
	};

	// float-value keyframe data
	typedef KeyFrameValueType<float> FloatValue;
	typedef std::set<FloatValue> FloatValueSet;


	// [RMS: this is very bad. A parameter could be deleted and the memory re-used, so the
	//   pointer is the same.  Parameters should be assigned a unique ID on creation! ]
	//! list of all float parameters in system
	std::map<rmsui::IParameter *, FloatValueSet> m_vFloatParameters;

	// Frame3f keyframe data
	typedef KeyFrameValueType<Wml::Frame3f> Frame3fValue;
	typedef std::set<Frame3fValue> Frame3fValueSet;

	// [RMS: see above. bad!]
	std::map<rmsui::SceneObject *, Frame3fValueSet> m_vFrame3fParameters;


	//! synchronize the internal list of parameters with the scenemanager
	void SynchronizeParameterList();

	//! set a keyframe for all parameters
	void SetKeyFrame( unsigned int nKeyFrame );

	//! remove all keys at a current frame
	void ClearKeyFrame( unsigned int nKeyFrame );

	// [RMS: using a float here so that we can increase the frame count. But maybe that is
	//  a bad idea. Should we maybe just scale the keyframe values instead... ?
	//! set the parameters for a keyframe value
	void SetParameters( float fTime );


	/*
	 * interpolation functions
	 */
	float InterpValueLinear( const FloatValue & vFirst, const FloatValue & vSecond, float fTime );
	Wml::Frame3f InterpFrameLinear( const Frame3fValue & vFirst, const Frame3fValue & vSecond, float fTime );

	Wml::Matrix3f FindRotation( const Wml::Frame3f & vCurFrame, const Wml::Frame3f & vDestFrame );

public:
	afx_msg void OnAnimationSetcurrentframe();
	afx_msg void OnAnimationSetlength();
	afx_msg void OnAnimationResetanimation();
	afx_msg void OnAnimationSetkeyframe();
	afx_msg void OnAnimationClearkeyframe();
};


