// Author: Ryan Schmidt  (rms@unknownroad.com)  http://www.unknownroad.com
// Copyright (c) 2006. All Rights Reserved.
// The SketchFramework Library is supplied "AS IS". The Author disclaims all warranties, expressed or implied, including, 
// without limitation, the warranties of merchantability and of fitness for any purpose. The Author assume no liability for 
// direct, indirect, incidental, special, exemplary, or consequential damages, which may result from the use of the 
// SketchFramework Library, even if advised of the possibility of such damage. Permission is hereby granted to use, copy, 
// modify, and distribute this source code, or portions hereof, for any purpose, without fee, subject to the following restrictions:
// 1) The origin of this source code must not be misrepresented.
// 2) This Copyright notice may not be removed or altered from any source or altered source distribution.
// The Author specifically permits, without fee, the use of this source code as a component in commercial products.

#ifndef __RMS_SPHERE_UTILS_H__
#define __RMS_SPHERE_UTILS_H__

#include <WmlVector3.h>
#include <vector>
#include "TriangleMesh.h"

namespace rmsimplicit {

class SphereUtils
{
public:

	// assumes both points are on sphere!
	static float GeodesicDistance( const Wml::Vector3f & vPoint1, const Wml::Vector3f & vPoint2 )
	{
		Wml::Vector3f fNorm1 = vPoint1;
		fNorm1.Normalize();
		
		Wml::Vector3f fNorm2 = vPoint2;
		fNorm2.Normalize();

		float fDot = fNorm1.Dot(fNorm2);
		if ( fDot < -1 ) fDot = -1;
		if ( fDot > 1 ) fDot = 1;
		float fAngle = acos( fDot );

		return vPoint1.Length() * fAngle;
	}


	static float GeodesicPolarAngle( const Wml::Vector3f & vCenter, const Wml::Vector3f & vPoint )
	{
		// grab normal at vCenter (easy!)
		Wml::Vector3f vNormal( vCenter );
		vNormal.Normalize();
		
		// compute transformation that takes vNormal to be aligned w/ Z axis
		Wml::Matrix3f matAlign;
		Wml::ComputeAlignZAxisMatrix( vNormal, matAlign, true );

		// ok, figure out vector to vPoint from vCenter, then transform it
		Wml::Vector3f vDirection( vPoint - vCenter );
		Wml::Vector3f vCanonDirection = matAlign * vDirection;

		// now we can just tack this vector onto the 'top' point of the sphere at (0, radius)

		// now we can compute the 'theta' from spherical coordinates
		float fTheta = atan2( vCanonDirection.Y() , vCanonDirection.X() );
		if ( ! _finite(fTheta) )
			fTheta = 0.0f;	

		// voila!
		return fTheta;
	}


	// returned as {r, theta, phi}, where theta in [0,2pi] and phi in [0,pi]
	// theta is longitude or azimuthal angle
	// phi is polar angle or colatitude (90 - latitude)
	static Wml::Vector3f GetSphericalCoords( const Wml::Vector3f & vXYZ ) 
	{
		float fR = vXYZ.Length();
		float fTheta = atan2( vXYZ.Y() , vXYZ.X() );
		if ( ! _finite(fTheta) ) {
			fTheta = 0.0f;
		}
		float fVal = vXYZ.Z() / fR;
		if ( fVal < -1 ) fVal = -1;
		if ( fVal > 1 ) fVal = 1;
		float fPhi = acos( fVal );
		return Wml::Vector3f( fR, fTheta, fPhi );
	}

	static Wml::Vector3f GetCartesianCoords( const Wml::Vector3f & vRTP )
	{
		float fX = vRTP.X() * cos( vRTP.Y() ) * sin( vRTP.Z() );
		float fY = vRTP.X() * sin( vRTP.Y() ) * sin( vRTP.Z() );
		float fZ = vRTP.X() * cos( vRTP.Z() );
		return Wml::Vector3f( fX, fY, fZ );
	}


	// returned as {r, long, lat}, where long in [0,2pi] and lat in [-pi/2,pi/2]
	static Wml::Vector3f GetLatLongCoords( const Wml::Vector3f & vXYZ ) 
	{
		float fR = vXYZ.Length();

		float fLong = atan2( vXYZ.Y() , vXYZ.X() );
		if ( ! _finite(fLong) ) {
			fLong = 0.0f;
		}
		float fVal = vXYZ.Z() / fR;
		if ( fVal < -1 ) fVal = -1;
		if ( fVal > 1 ) fVal = 1;
		float fLat = Wml::Mathf::HALF_PI - acos( fVal );

		return Wml::Vector3f( fR, fLong, fLat );
	}

	// return value is (r,Theta)
	static Wml::Vector2f GetPolarGeodesicCoords( const Wml::Vector3f & vSeed, const Wml::Vector3f & vPoint )
	{
		// ok, compute geodesic distance
		float fDist = GeodesicDistance( vSeed, vPoint );

		// figure out geodesic angle
		float fAngle = GeodesicPolarAngle( vSeed, vPoint );

		return Wml::Vector2f( fDist, fAngle );
	}

	// return is (u,v)
	static Wml::Vector2f GetNormalCoords( const Wml::Vector3f & vSeed, const Wml::Vector3f & vPoint )
	{
		// ok, compute geodesic distance
		float fDist = GeodesicDistance( vSeed, vPoint );

		// figure out geodesic angle
		float fAngle = GeodesicPolarAngle( vSeed, vPoint );

		// convert (r,T) polar coords to (u,v)
		return Wml::Vector2f( fDist * cos( fAngle ), fDist * sin( fAngle ) );
	}


	static void ParamByExponentialMap( TriangleMesh & mesh, const Wml::Vector3f & vSeedPoint )
	{
		unsigned int nVerts = mesh.GetNumVertices();
		for ( unsigned int i = 0; i < nVerts; ++i ) {
			Wml::Vector3f vVert;
			mesh.GetVertex(i, vVert);

			// ok, compute geodesic distance
			float fDist = GeodesicDistance( vSeedPoint, vVert );

			// figure out geodesic angle
			float fAngle = GeodesicPolarAngle( vSeedPoint, vVert );

			// convert (r,T) polar coords to (u,v)
			Wml::Vector2f vCoord( fDist * cos( fAngle ), fDist * sin( fAngle ) );

			mesh.AddVertexData(i, NULL, NULL, NULL, vCoord);
		}

	}

	static void ParamBySphericalCoords( TriangleMesh & mesh )
	{
		unsigned int nVerts = mesh.GetNumVertices();
		for ( unsigned int i = 0; i < nVerts; ++i ) {
			Wml::Vector3f vVert;
			mesh.GetVertex(i, vVert);

			Wml::Vector3f vSpherical = SphereUtils::GetSphericalCoords( vVert );

			Wml::Vector2f vCoord( 
				vSpherical.Y() / Wml::Mathf::TWO_PI,
				vSpherical.Z() / Wml::Mathf::PI );

			mesh.AddVertexData(i, NULL, NULL, NULL, vCoord);
		}
	}



	static void ProjectMeshToSphere( TriangleMesh & mesh, float fRadius )
	{
		unsigned int nVerts = mesh.GetNumVertices();
		for ( unsigned int i = 0; i < nVerts; ++i ) {
			Wml::Vector3f vVert;
			mesh.GetVertex(i, vVert);
			vVert.Normalize();
			vVert *= fRadius;

			mesh.SetVertexData(i, vVert);
		}
	}


	static void CreateRandomPointDistrib( unsigned int nPoints, TriangleMesh & mesh )
	{
		mesh.Clear();

		srand(13413131);

		for ( unsigned int i = 0; i < nPoints; ++i ) {

			float fX = 0.5f - (float)rand() / (float)RAND_MAX;
			float fY = 500.0f - (float)rand() / (float)RAND_MAX;
			float fZ = 500.0f - (float)rand() / (float)RAND_MAX;

			Wml::Vector3f vPoint( fX, fY, fZ );
			vPoint.Normalize();

			mesh.AppendVertexData( vPoint, vPoint, false );
		}
	}


	static void CreateRegularSphereMesh( unsigned int nNumSegments, unsigned int nNumSlices,
										 float fRadius, TriangleMesh & mesh )
	{
		mesh.Clear();

		// generate bottom vertex
		Wml::Vector3f vBottom(0.0f, 0.0f, -fRadius);
		Wml::Vector3f vBottomNormal(0.0f, 0.0f, -1.0f);

		unsigned int nBottomVertexIndex = mesh.AppendVertexData(&vBottom, &vBottomNormal, false );

		unsigned int nSlice = 1;
		float fSliceThetaInc = Wml::Mathf::PI / (nNumSlices-1);
		float fSegmentThetaInc = Wml::Mathf::TWO_PI / nNumSegments;

		// generate first slice
		float fLatitude = -Wml::Mathf::HALF_PI + (nSlice * fSliceThetaInc);
		float fLongitude = 0.0;
		unsigned int nLastVertex = 0;
		for ( unsigned int nSegment = 0; nSegment < nNumSegments; ++nSegment ) {
			
			fLongitude = nSegment * fSegmentThetaInc;

			Wml::Vector3f R1(
				fRadius * cos(fLongitude) * cos(fLatitude),
				fRadius * sin(fLongitude) * cos(fLatitude),
				fRadius * sin(fLatitude) );

			Wml::Vector3f R1Normal( R1 );
			R1Normal.Normalize();

			unsigned int nNewVertex = mesh.AppendVertexData( &R1, &R1Normal, false );

			if ( nSegment > 0 ) {
				unsigned int nTriangle[3] = {nBottomVertexIndex, nNewVertex, nLastVertex};
				mesh.AppendTriangleData(nTriangle);
			}

			if ( nSegment == nNumSegments-1 ) {
				unsigned int nTriangle[3] = {nBottomVertexIndex, nNewVertex-nNumSegments+1, nNewVertex};
				mesh.AppendTriangleData(nTriangle);
			}

			nLastVertex = nNewVertex;
		}


		for ( ; nSlice < nNumSlices-1; ++nSlice ) {

			fLatitude = -Wml::Mathf::HALF_PI + (nSlice * fSliceThetaInc);
			fLongitude = 0.0;
			unsigned int nLastSliceVertex = nLastVertex-nNumSegments+1;
			nLastVertex = 0;
			for ( unsigned int nSegment = 0; nSegment < nNumSegments; ++nSegment ) {
				
				fLongitude = nSegment * fSegmentThetaInc;

				Wml::Vector3f R1(
					fRadius * cos(fLongitude) * cos(fLatitude),
					fRadius * sin(fLongitude) * cos(fLatitude),
					fRadius * sin(fLatitude) );

				Wml::Vector3f R1Normal( R1 );
				R1Normal.Normalize();

				unsigned int nNewVertex = mesh.AppendVertexData( &R1, &R1Normal, false );

				if ( nSegment > 0 ) {
					unsigned int nTriangle1[3] = { nLastSliceVertex + nSegment - 1, nNewVertex, nLastVertex };
					mesh.AppendTriangleData(nTriangle1);
					unsigned int nTriangle2[3] = { nLastSliceVertex + nSegment - 1, nLastSliceVertex + nSegment, nNewVertex};
					mesh.AppendTriangleData(nTriangle2);
				}

				if ( nSegment == nNumSegments-1 ) {
					unsigned int nTriangle1[3] = { nLastSliceVertex + nSegment, nNewVertex-nNumSegments+1, nNewVertex};
					mesh.AppendTriangleData(nTriangle1);
					unsigned int nTriangle2[3] = { nLastSliceVertex + nSegment, nLastSliceVertex, nNewVertex-nNumSegments+1};
					mesh.AppendTriangleData(nTriangle2);
				}

				nLastVertex = nNewVertex;
			}

		}


		// generate top vertex
		Wml::Vector3f vTop(0.0f, 0.0f, fRadius);
		Wml::Vector3f vTopNormal(0.0f, 0.0f, 1.0f);
		unsigned int nTopVertexIndex = mesh.AppendVertexData(&vTop, &vTopNormal, false );

		// generate first slice
		fLatitude = -Wml::Mathf::HALF_PI + (nSlice * fSliceThetaInc);
		fLongitude = 0.0;
		unsigned int nLastSliceVertex = nLastVertex-nNumSegments+1;
		for ( unsigned int nSegment = 0; nSegment < nNumSegments; ++nSegment ) {
			
			fLongitude = nSegment * fSegmentThetaInc;

			Wml::Vector3f R1(
				fRadius * cos(fLongitude) * cos(fLatitude),
				fRadius * sin(fLongitude) * cos(fLatitude),
				fRadius * sin(fLatitude) );

			Wml::Vector3f R1Normal( R1 );
			R1Normal.Normalize();

			unsigned int nNewVertex = nLastSliceVertex + nSegment;

			if ( nSegment > 0 ) {
				unsigned int nTriangle[3] = {nTopVertexIndex, nLastVertex, nNewVertex};
				mesh.AppendTriangleData(nTriangle);
			}

			if ( nSegment == nNumSegments-1 ) {
				unsigned int nTriangle[3] = {nTopVertexIndex, nNewVertex, nNewVertex-nNumSegments+1};
				mesh.AppendTriangleData(nTriangle);
			}

			nLastVertex = nNewVertex;
		}

	}




};


} // end namespace Wml

#endif // __RMS_SPHERE_UTILS_H__