// Ryan Schmidt - rms@unknownroad.com
// Copyright (c) 2004-2010
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt


#include "StdAfx.h"
#include "VFTriangleMesh.h"

#include <GL/gl.h>
#include <GL/glu.h>

#include "rmsdebug.h"
#include "rmsprofile.h"

using namespace rmsimplicit;


// these warnings are about type casting between pointers and uints
#pragma warning( disable : 4311 )
#pragma warning( disable : 4312 )


const VFTriangleMesh::VertexID VFTriangleMesh::INVALID_VERTEX = 0;
const VFTriangleMesh::VertexID VFTriangleMesh::INVALID_NORMAL = 0;
const VFTriangleMesh::VertexID VFTriangleMesh::INVALID_TRIANGLE = 0;

VFTriangleMesh::VFTriangleMesh()
{
	// implementation depends on this!!!
	ASSERT( sizeof(VFTriangle *) == sizeof( unsigned int ) );
}

VFTriangleMesh::~VFTriangleMesh()
{
}

VFTriangleMesh::VertexID VFTriangleMesh::AddVertex( const Wml::Vector3f & vVertex )
{
	VFVertex * pVertex = m_VertexPool.Allocate();
	pVertex->v = vVertex;
	pVertex->nRefCount = 0;
	pVertex->vTriangles.clear();

	VertexID id = (unsigned int)pVertex;
	m_vVertices[id] = pVertex;
	return id;
}

VFTriangleMesh::NormalID VFTriangleMesh::AddNormal( const Wml::Vector3f & vNormal )
{
	VFNormal * pNormal = m_NormalPool.Allocate();
	pNormal->n = vNormal;
	pNormal->nRefCount = 0;

	NormalID id = (unsigned int)pNormal;
	m_vNormals[id] = pNormal;
	return id;
}

VFTriangleMesh::TriangleID VFTriangleMesh::AddTriangle( VFTriangleMesh::VertexID v1, 
													    VFTriangleMesh::VertexID v2, 
														VFTriangleMesh::VertexID v3, 
														VFTriangleMesh::NormalID n1, 
														VFTriangleMesh::NormalID n2, 
														VFTriangleMesh::NormalID n3 )
{
	VFTriangle * pTriangle = m_TrianglePool.Allocate();

	ASSERT( m_vVertices.find(v1) != m_vVertices.end() );
	ASSERT( m_vVertices.find(v2) != m_vVertices.end() );
	ASSERT( m_vVertices.find(v3) != m_vVertices.end() );
	ASSERT( n1 == INVALID_NORMAL || m_vNormals.find(n1) != m_vNormals.end() );
	ASSERT( n2 == INVALID_NORMAL || m_vNormals.find(n2) != m_vNormals.end() );
	ASSERT( n3 == INVALID_NORMAL || m_vNormals.find(n3) != m_vNormals.end() );

	TriangleID id = (unsigned int)pTriangle;
	m_vTriangles[id] = pTriangle;

	pTriangle->vVertices[0] = v1;
	pTriangle->vVertices[1] = v2;
	pTriangle->vVertices[2] = v3;
	pTriangle->vNormals[0] = n1;
	pTriangle->vNormals[1] = n2;
	pTriangle->vNormals[2] = n3;

	((VFVertex *)v1)->nRefCount++;
	((VFVertex *)v2)->nRefCount++;
	((VFVertex *)v3)->nRefCount++;

	((VFVertex *)v1)->vTriangles.insert(id);
	((VFVertex *)v2)->vTriangles.insert(id);
	((VFVertex *)v3)->vTriangles.insert(id);

	// increment normal refcounts
	if (n1 != INVALID_NORMAL)
		((VFNormal *)n1)->nRefCount++;
	if (n2 != INVALID_NORMAL)
		((VFNormal *)n2)->nRefCount++;
	if (n3 != INVALID_NORMAL)
		((VFNormal *)n3)->nRefCount++;


	return id;
}


bool VFTriangleMesh::RemoveVertex( VFTriangleMesh::VertexID id )		// removes attached triangles and normals
{
	std::map< VertexID, VFVertex * >::iterator itr( m_vVertices.find(id) );
	if ( itr == m_vVertices.end() )
		return false;				// ignore bad removes

	VFVertex * pVertex = itr->second;

	// ok now remove each attached face

	// need to make a list here because RemoveTriangle will mess up iterator...
	//m_vTempTriangles.resize( 0 );
	//std::vector<TriangleID> triangles;
	//std::set< TriangleID >::iterator cur( pVertex->vTriangles.begin()), end( pVertex->vTriangles.end() );
	//while ( cur != end )
	//	triangles.push_back( *cur++ );

	//size_t nCount = triangles.size();
	//for ( unsigned int i = 0; i < nCount; ++i)
	//	RemoveTriangle( triangles[i] );

	while (! pVertex->vTriangles.empty() )
		RemoveTriangle( pVertex->vTriangles.begin() );

	// vertex should be gone now because no attached faces remain! 
	ASSERT( m_vVertices.find(id) == m_vVertices.end() );

	return true;
}


bool VFTriangleMesh::RemoveTriangle( VFTriangleMesh::TriangleID id )
{
	std::map< TriangleID, VFTriangle * >::iterator itr( m_vTriangles.find(id) );
	if ( itr == m_vTriangles.end() )
		return false;						// ignore bad removes

	VFTriangle * pTriangle = itr->second;

	// decrement vertex refcounts. remove any with refcount = 0;
	for ( unsigned int i = 0; i < 3; ++i ) {
		if ( pTriangle->vVertices[i] != INVALID_VERTEX ) {
			VFVertex * pVertex = (VFVertex *)pTriangle->vVertices[i];
			pVertex->vTriangles.erase( id );
			if ( --pVertex->nRefCount == 0 )
				DeleteVertex( pTriangle->vVertices[i] );
		}
	}

	// decrement normal refcounts. remove any with refcount = 0;
	for ( unsigned int i = 0; i < 3; ++i ) {
		if ( pTriangle->vNormals[i] != INVALID_NORMAL ) {
			VFNormal * pNormal = (VFNormal *)pTriangle->vNormals[i];
			if ( --pNormal->nRefCount == 0 )
				DeleteNormal( pTriangle->vNormals[i] );
		}
	}

	// kill triangle
//	DeleteTriangle( id );
	DeleteTriangle( id, pTriangle );

	// make sure triangle is gone...
	ASSERT( m_vTriangles.find(id) == m_vTriangles.end() );

	return true;
}



bool VFTriangleMesh::DeleteVertex( VFTriangleMesh::VertexID id )
{
	if ( id == INVALID_VERTEX )
		return false;

	if ( m_vVertices.find(id) != m_vVertices.end() ) {
		((VFVertex *)id)->vTriangles.clear();
		m_vVertices.erase(id);
		m_VertexPool.Free( (VFVertex *)id );
	}
	return true;
}

bool VFTriangleMesh::DeleteNormal( VFTriangleMesh::NormalID id )
{
	if ( id == INVALID_NORMAL )
		return false;
	
	if ( m_vNormals.find(id) != m_vNormals.end() ) {
		m_vNormals.erase(id);
		m_NormalPool.Free( (VFNormal *)id );
	}
	return true;
}

bool VFTriangleMesh::DeleteTriangle( VFTriangleMesh::TriangleID id )
{
	if ( id == INVALID_TRIANGLE )
		return false;

	if ( m_vTriangles.find(id) != m_vTriangles.end() ) {
		m_vTriangles.erase(id);
		m_TrianglePool.Free( (VFTriangle *)id );
	}
	return true;
}

void VFTriangleMesh::DeleteTriangle( TriangleID id, VFTriangle * pTriangle )
{
	m_vTriangles.erase( id );
	m_TrianglePool.Free( pTriangle );
}




void VFTriangleMesh::Clear()
{
	m_vVertices.clear();
	m_vNormals.clear();
	m_vTriangles.clear();

	m_VertexPool.ClearAll();
	m_NormalPool.ClearAll();
	m_TrianglePool.ClearAll();
}


unsigned int VFTriangleMesh::GetVertexCount()
{
	return (unsigned int)m_vVertices.size();
}

unsigned int VFTriangleMesh::GetNormalCount()
{
	return (unsigned int)m_vNormals.size();
}

unsigned int VFTriangleMesh::GetTriangleCount()
{
	return (unsigned int)m_vTriangles.size();
}


/*
 * RMS TODO: inline all this!! bloody hell...
 */


Wml::Vector3f VFTriangleMesh::GetVertex( VFTriangleMesh::VertexID id )
{	
	ASSERT( m_vVertices.find(id) != m_vVertices.end() );
	return ((VFVertex *)id)->v;
}

void VFTriangleMesh::GetVertex( VFTriangleMesh::VertexID id, Wml::Vector3f & vVertex )
{
	ASSERT( m_vVertices.find(id) != m_vVertices.end() );
	vVertex = ((VFVertex *)id)->v;
}

Wml::Vector3f VFTriangleMesh::GetNormal( VFTriangleMesh::NormalID id )
{
	ASSERT( m_vNormals.find(id) != m_vNormals.end() );
	return ((VFNormal *)id)->n;
}

void VFTriangleMesh::GetNormal( VFTriangleMesh::NormalID id, Wml::Vector3f & vNormal )
{
	ASSERT( m_vNormals.find(id) != m_vNormals.end() );
	vNormal = ((VFNormal *)id)->n;
}

VFTriangleMesh::NormalID VFTriangleMesh::GetAnyNormal( VFTriangleMesh::VertexID id, Wml::Vector3f & vNormal )
{
	ASSERT( m_vVertices.find(id) != m_vVertices.end() );
	SmallTriangleList & vTriangles = ((VFVertex *)id)->vTriangles;
//	std::set< TriangleID > * pTriangles = & ((VFVertex *)id)->vTriangles;
//	if ( ! pTriangles->empty() ) {
	if ( ! vTriangles.empty() ) {
		TriangleID tID = vTriangles.begin();
		for ( unsigned int i = 0; i < 3; ++i ) {
			if ( GetVertexID( tID, i ) == id ) {
				GetNormal( tID, i, vNormal );
				return GetNormalID( tID, i );
			}
		}
	}
	vNormal = Wml::Vector3f::UNIT_Z;
	return INVALID_NORMAL;
}

VFTriangleMesh::VertexID VFTriangleMesh::GetVertexID( VFTriangleMesh::TriangleID id, unsigned int nVertex )
{
	ASSERT( m_vTriangles.find(id) != m_vTriangles.end() );
	return ((VFTriangle *)id)->vVertices[nVertex];
}

VFTriangleMesh::NormalID VFTriangleMesh::GetNormalID( VFTriangleMesh::TriangleID id, unsigned int nVertex )
{
	ASSERT( m_vTriangles.find(id) != m_vTriangles.end() );
	return ((VFTriangle *)id)->vNormals[nVertex];
}

Wml::Vector3f VFTriangleMesh::GetVertex( VFTriangleMesh::TriangleID id, unsigned int nVertex )
{
	ASSERT( m_vTriangles.find(id) != m_vTriangles.end() );
	VertexID vID = ((VFTriangle *)id)->vVertices[nVertex];
	ASSERT( m_vVertices.find(vID) != m_vVertices.end() );
	return ((VFVertex *)vID)->v;
}

VFTriangleMesh::VertexID VFTriangleMesh::GetVertex( VFTriangleMesh::TriangleID id, unsigned int nVertex, Wml::Vector3f & vVertex )
{
	ASSERT( m_vTriangles.find(id) != m_vTriangles.end() );
	VertexID vID = ((VFTriangle *)id)->vVertices[nVertex];
	ASSERT( m_vVertices.find(vID) != m_vVertices.end() );
	vVertex = ((VFVertex *)vID)->v;
	return vID;
}

Wml::Vector3f VFTriangleMesh::GetNormal( VFTriangleMesh::TriangleID id, unsigned int nVertex )
{
	ASSERT( m_vTriangles.find(id) != m_vTriangles.end() );
	NormalID nID = ((VFTriangle *)id)->vNormals[nVertex];
	ASSERT( m_vNormals.find(nID) != m_vNormals.end() );
	return ((VFNormal *)nID)->n;
}

void VFTriangleMesh::GetNormal( VFTriangleMesh::TriangleID id, unsigned int nVertex, Wml::Vector3f & vNormal )
{
	ASSERT( m_vTriangles.find(id) != m_vTriangles.end() );
	NormalID nID = ((VFTriangle *)id)->vNormals[nVertex];
	ASSERT( m_vNormals.find(nID) != m_vNormals.end() );
	vNormal = ((VFNormal *)nID)->n;
}

VFTriangleMesh::VertexID VFTriangleMesh::GetVertex( VFTriangleMesh::TriangleID id, unsigned int nVertex, Wml::Vector3f & vVertex, Wml::Vector3f & vNormal )
{
	ASSERT( m_vTriangles.find(id) != m_vTriangles.end() );
	VertexID vID = ((VFTriangle *)id)->vVertices[nVertex];
	NormalID nID = ((VFTriangle *)id)->vNormals[nVertex];
	ASSERT( m_vVertices.find(vID) != m_vVertices.end() );
	vVertex = ((VFVertex *)vID)->v;
	ASSERT( m_vNormals.find(nID) != m_vNormals.end() );
	vNormal = ((VFNormal *)nID)->n;
	return vID;
}


void VFTriangleMesh::GetTriangle( TriangleID id, Wml::Vector3f * pStore )
{
	ASSERT( m_vTriangles.find(id) != m_vTriangles.end() );
	for ( int i = 0; i < 3; ++i ) {
		VertexID vID = ((VFTriangle *)id)->vVertices[i];
		ASSERT( m_vVertices.find(vID) != m_vVertices.end() );
		pStore[i] = ((VFVertex *)vID)->v;
	}
}

void VFTriangleMesh::GetTriangle( TriangleID id, VertexID * pStore )
{
	ASSERT( m_vTriangles.find(id) != m_vTriangles.end() );
	for ( int i = 0; i < 3; ++i ) {
		pStore[i] = ((VFTriangle *)id)->vVertices[i];
	}
}

void VFTriangleMesh::GetTriangleNormals( TriangleID id, Wml::Vector3f * pStore )
{
	ASSERT( m_vTriangles.find(id) != m_vTriangles.end() );
	for ( int i = 0; i < 3; ++i ) {
		NormalID nID = ((VFTriangle *)id)->vNormals[i];
		ASSERT( m_vNormals.find(nID) != m_vNormals.end() );
		pStore[i] = ((VFNormal *)nID)->n;
	}
}

void VFTriangleMesh::GetTriangleNormals( TriangleID id, NormalID * pStore )
{
	ASSERT( m_vTriangles.find(id) != m_vTriangles.end() );
	for ( int i = 0; i < 3; ++i ) {
		pStore[i] = ((VFTriangle *)id)->vNormals[i];
	}
}

VFTriangleMesh::TriangleID VFTriangleMesh::GetEdgeNeighbour( VFTriangleMesh::TriangleID id, int v1idx, int v2idx )
{
	ASSERT( m_vTriangles.find(id) != m_vTriangles.end() );

	VertexID v1 = ((VFTriangle *)id)->vVertices[v1idx];
	VertexID v2 = ((VFTriangle *)id)->vVertices[v2idx];

	// ok, loop through triangles of v1. One of them should have a v1/v2 edge, that one is our neighbour
	//std::set< TriangleID >::iterator cur(((VFVertex *)v1)->vTriangles.begin()), 
	//	end( ((VFVertex *)v1)->vTriangles.end());
	//while ( cur != end ) {
	//	TriangleID tID = (*cur++);
	unsigned int nCount = ((VFVertex *)v1)->vTriangles.size();
	for ( unsigned int i = 0; i < nCount; ++i ) {
		TriangleID tID = ((VFVertex *)v1)->vTriangles.get(i);

		if ( tID == id )
			continue;

		VertexID * pVertices = ((VFTriangle *)tID)->vVertices;
		for ( int i = 0; i < 3; ++i ) {
			if ( pVertices[i] == v1 ) {
				if ( pVertices[ (i+1)%3 ] == v2 || pVertices[ (i+2)%3 ] == v2 )
					return tID;
			}
		}
	}

	return INVALID_TRIANGLE;
}




void VFTriangleMesh::SetVertex( VFTriangleMesh::VertexID id, const Wml::Vector3f & vVertex )
{
	ASSERT( m_vVertices.find(id) != m_vVertices.end() );
	((VFVertex *)id)->v = vVertex;
}


void VFTriangleMesh::SetNormal( VFTriangleMesh::NormalID id, const Wml::Vector3f & vNormal )
{
	ASSERT( m_vNormals.find(id) != m_vNormals.end() );
	((VFNormal *)id)->n = vNormal;
}


unsigned int VFTriangleMesh::GetTriangleCount( VFTriangleMesh::VertexID id )
{
	ASSERT( m_vVertices.find(id) != m_vVertices.end() );
	return (unsigned int)( ((VFVertex *)id)->vTriangles.size() );
}


bool VFTriangleMesh::GetNextTriangle( VertexID id, TriangleID & lastTriangle )
{
	ASSERT( m_vVertices.find(id) != m_vVertices.end() );

	SmallTriangleList & vTriangles = ((VFVertex *)id)->vTriangles;
	if ( lastTriangle == VFTriangleMesh::INVALID_TRIANGLE ) {
		if ( vTriangles.empty() )
			return false;
		else {
			lastTriangle = vTriangles.begin();
			return true;
		}
	} else {
		unsigned int nCount = vTriangles.size();
		TriangleID nextTriangle = VFTriangleMesh::INVALID_TRIANGLE;
		for ( unsigned int i = 0; i < nCount && nextTriangle == VFTriangleMesh::INVALID_TRIANGLE; ++i ) {
			if ( vTriangles.get(i) == lastTriangle )
				nextTriangle = vTriangles.get(i+1);
		}
		if ( nextTriangle == VFTriangleMesh::INVALID_TRIANGLE)
			return false;
		lastTriangle = nextTriangle;
	}

	return true;
}


bool VFTriangleMesh::GetUnordered1Ring( VertexID id, std::vector<VertexID> & vStore )
{
	ASSERT(0 && "BORKEN");
	return false;
#if 0
	// [RMS: we could optimize this a lot of the lists of triangles were ordered
	//   at each vertex!]

	ASSERT( m_vVertices.find(id) != m_vVertices.end() );

	size_t nMaxCount = vStore.capacity();
	if ( vStore.size() == nMaxCount )
		return false;

	std::set< TriangleID > * pTriangles = & ((VFVertex *)id)->vTriangles;
	std::set< TriangleID >::iterator cur( pTriangles->begin() ), end( pTriangles->end() );
	while ( cur != end ) {

		size_t nCurCount = vStore.size();
		
		TriangleID tID = (*cur);

		for ( unsigned int i = 0; i  < 3; ++i ) {

			VertexID tVID = GetVertexID( tID, i );
			if ( tVID == id )
				continue;

			bool bFound = false;
			for ( unsigned int j = 0; j < nCurCount && !bFound; ++j ) {
				if ( vStore[j] == tVID )
					bFound = true;
			}

			if (!bFound) {

				vStore.push_back( tVID );
				size_t nCurCount = vStore.size();
				if ( nCurCount == nMaxCount )
					return false;
			}
		}
		
		cur++;
	}

	return true;
#endif
}


void VFTriangleMesh::DrawGL(bool bNormals)
{
	glBegin(GL_TRIANGLES);
	std::map< TriangleID, VFTriangle * >::iterator cur( m_vTriangles.begin()), end ( m_vTriangles.end() );
	if (bNormals) {
		while (cur != end) {
			VFTriangle * pTriangle = (*cur++).second;
			VFVertex * p1 = (VFVertex *)pTriangle->vVertices[0];
			VFVertex * p2 = (VFVertex *)pTriangle->vVertices[1];
			VFVertex * p3 = (VFVertex *)pTriangle->vVertices[2];
			glNormal3fv( ((VFNormal *)pTriangle->vNormals[0])->n );
			glVertex3fv( ((VFVertex *)pTriangle->vVertices[0])->v );
			glNormal3fv( ((VFNormal *)pTriangle->vNormals[1])->n );
			glVertex3fv( ((VFVertex *)pTriangle->vVertices[1])->v );
			glNormal3fv( ((VFNormal *)pTriangle->vNormals[2])->n );
			glVertex3fv( ((VFVertex *)pTriangle->vVertices[2])->v );
		}
	} else {
		while (cur != end) {
			VFTriangle * pTriangle = (*cur++).second;
			glVertex3fv( ((VFVertex *)pTriangle->vVertices[0])->v );
			glVertex3fv( ((VFVertex *)pTriangle->vVertices[1])->v );
			glVertex3fv( ((VFVertex *)pTriangle->vVertices[2])->v );
		}
	}
	glEnd();
}



bool VFTriangleMesh::InitializeFromSimpleMesh( rmsimplicit::TriangleMesh * pMesh )
{
	std::vector<VertexID> vVertexIDs;
	std::vector<NormalID> vNormalIDs;

	size_t nVCount = pMesh->GetNumVertices();
	vVertexIDs.resize( nVCount );
	vNormalIDs.resize( nVCount );

	for ( unsigned int i = 0; i < nVCount; ++i ) {
		Wml::Vector3f v;
		pMesh->GetVertex( i, v );
		vVertexIDs[i] = AddVertex( v );
		pMesh->GetNormal( i, v );
		vNormalIDs[i] = AddNormal( v );
	}

	size_t nTCount = pMesh->GetNumTriangles();
	unsigned int nTriangle[3];
	for ( unsigned int i = 0; i < nTCount; ++i ) {
		pMesh->GetTriangle( i, nTriangle );

		AddTriangle( vVertexIDs[ nTriangle[0] ], vVertexIDs[ nTriangle[1] ], vVertexIDs[ nTriangle[2] ],
			vNormalIDs[ nTriangle[0] ], vNormalIDs[ nTriangle[1] ], vNormalIDs[ nTriangle[2] ] );
	}

	return true;
}



// SmallTriangleList class

void VFTriangleMesh::SmallTriangleList::insert( TriangleID id )
{
	if ( m_vTriangles.size() == m_vTriangles.capacity() )
		m_vTriangles.reserve( m_vTriangles.capacity() + 2 );

	if ( m_vTriangles[0] == INVALID_TRIANGLE ) {
		m_vTriangles.resize( m_vTriangles.size() + 1 );
		m_vTriangles[0] = id;
		m_vTriangles[1] = INVALID_TRIANGLE;
	} else {
		int i = 0;
		while ( id > m_vTriangles[i] && m_vTriangles[i] != INVALID_TRIANGLE ) 
			++i;

		if ( m_vTriangles[i] != id ) {

			m_vTriangles.resize( m_vTriangles.size() + 1 );

			int j = i;
			if ( m_vTriangles[j] == INVALID_TRIANGLE ) {
				m_vTriangles[j] = id;
				m_vTriangles[j+1] = INVALID_TRIANGLE;
			} else {
				TriangleID tmp1 = m_vTriangles[j], tmpSwap;
				++j;
				for ( ; m_vTriangles[j] != INVALID_TRIANGLE; ++j ) {
					tmpSwap = m_vTriangles[j];
					m_vTriangles[j] = tmp1;
					tmp1 = tmpSwap;
				}
				m_vTriangles[j] = tmp1;
				m_vTriangles[j+1] = INVALID_TRIANGLE;
				m_vTriangles[i] = id;
			}

		}
	}
}

bool VFTriangleMesh::SmallTriangleList::find( TriangleID id )
{
	return false;
}


void VFTriangleMesh::SmallTriangleList::erase( TriangleID id )
{
	// find id
	int i = 0;
	while ( m_vTriangles[i] != id && m_vTriangles[i] != INVALID_TRIANGLE )
		++i;

	if ( m_vTriangles[i] == INVALID_TRIANGLE )
		return;

	int j = i;
	while( m_vTriangles[j] != INVALID_TRIANGLE ) {	// will wrap, but that's ok because we copied INVALID_TRIANGLE !!
		m_vTriangles[j] = m_vTriangles[j+1];
		j++;
	}

	m_vTriangles.resize( m_vTriangles.size() - 1 );
}




#pragma warning( default : 4311 )
#pragma warning( default : 4312 )
