// Ryan Schmidt - rms@unknownroad.com
// Copyright (c) 2005-2010
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt


#include "StdAfx.h"
#include "Geometry.h"

#include <gl/GL.h>
#include <gl/GLU.h>
#include <WmlMatrix3.h>
#include <WmlIntrLin3Sph3.h>
#include <WmlExtDistLin3Sph3.h>
#include <WmlDistLin3Cir3.h>

using namespace rmsui;


/*!
 * ShapeManager lets us share display lists between objects and still be HGLRC-safe...
 * @todo: make more thread safe ?? TODO
 */
class ShapeManager
{
public:
	//! @enum supported shape types
	enum ShapeType
	{
		Sphere = 0,
		Circle,
		Chevron
	};

	ShapeManager( HGLRC hGLRC )
		: m_hGLRC(hGLRC)
		{ memset(m_displayLists, 0, 25 * sizeof(unsigned int)); }	


	//! Get OpenGL display list ID for given shape
	unsigned int GetDisplayListByShape( ShapeType eType )
	{
		if (m_displayLists[(int)eType] == 0) {
			m_displayLists[(int)eType] = glGenLists(1);
			glNewList(m_displayLists[(int)eType], GL_COMPILE);
			GenerateShape(eType);
			glEndList();
		}
		return m_displayLists[(int)eType];
	}

	//! Get shape manager for current GLRC. Returns NULL if no GLRC is current (!)
	static ShapeManager * GetCurrent()
	{
		HGLRC hGLRC = wglGetCurrentContext();
		if (hGLRC == NULL) {
			ASSERT(0 && "NEED CURRENT GLRC!");
			return NULL;
		}

		ManagerMap::iterator found( s_managers.find(hGLRC) );
		if (found != s_managers.end()) {
			return found->second;
		}
		ShapeManager * pManager = new ShapeManager(hGLRC);
		s_managers.insert( std::pair<HGLRC, ShapeManager *>(hGLRC,pManager) );
		return pManager;
	}

protected:
	
	//! shape manager holds display lists for a current GLRC
	HGLRC m_hGLRC;

	//! display list indexes - hardcoded size (booerns!)
	unsigned int m_displayLists[25];

	void GenerateShape( ShapeType eType )
	{
		switch (eType) {
			case Sphere:
				{
					GLUquadric * quadric = gluNewQuadric();
					gluSphere( quadric, 1.0, 20, 20 );
					gluDeleteQuadric( quadric );
					break;
				}

			case Circle:
				{
					unsigned int nDivisions = 60;
					float fAngleInc = (2.0f*Wml::Mathf::PI) / ((float)nDivisions);
					glBegin(GL_LINE_LOOP);
					for (unsigned int i = 0; i < nDivisions; ++i) {
						float fAngle = fAngleInc * (float)i;
						float fCosAngle = Wml::Mathf::Cos(fAngle);
						float fSinAngle = Wml::Mathf::Sin(fAngle);
						glVertex3f(fCosAngle, fSinAngle, 0.0);
					}
					glEnd();
				}
				break;

				case Chevron:
				{
					glBegin(GL_LINE_STRIP);
					glVertex3f(-1, 0, 0);
					glVertex3f(0, 0.6f, 0);
					glVertex3f(1, 0, 0);
					glEnd();
				}
				break;	

			default:
				ASSERT(0 && "INVALID TYPE");
		}
	}

	typedef std::map<HGLRC, ShapeManager *> ManagerMap;

	//! use static map to get shape manager for current HGLRC
	static ManagerMap s_managers;
};

std::map<HGLRC, ShapeManager *> ShapeManager::s_managers;



/*
 * Sphere
 */

void Sphere::Render( const Wml::ColorRGBA & color )
{
	if (m_sphereDisplayList == 0) {
		m_sphereDisplayList = ShapeManager::GetCurrent()->GetDisplayListByShape( ShapeManager::Sphere );
	}

	glColor4f( color.R(), color.G(), color.B(), color.A() );
	glPushMatrix();
	glTranslatef( m_sphere.Center().X(), m_sphere.Center().Y(), m_sphere.Center().Z() );
	glScalef( m_sphere.Radius(), m_sphere.Radius(), m_sphere.Radius() );
	glCallList( m_sphereDisplayList );
	glPopMatrix();
}

bool Sphere::Intersect( const Wml::Ray3f & ray, 
					    Wml::Vector3f & ptPosition, float & fDepth )
{
	bool bResult = false;
	int iQuantity;
	Wml::Vector3f points[2];
	if (Wml::FindIntersection(ray, m_sphere, iQuantity, points)) {
		float fRayOrigin = ray.Direction().Dot(ray.Origin());
		float fNearest = Wml::Mathf::MAX_REAL;
		for (int i = 0; i < iQuantity; ++i) {
			float fPos = ray.Direction().Dot( points[i] );
			if (fPos > fRayOrigin && fPos < fNearest) {
				fNearest = fPos;
				fDepth = fPos;
				ptPosition = points[i];
			}
		}
		bResult = (fNearest < Wml::Mathf::MAX_REAL);
	} 

	return bResult;
}

void Sphere::ClosestPoint( const Wml::Line3f & line, Wml::Vector3f & ptPosition )
{
	Wml::Distance( line, m_sphere, (Wml::Vector3f *)NULL, &ptPosition);
}


/*
 * Circle
 */

const float Circle::s_IntersectionTolerance = 20.0f;

void Circle::Render( const Wml::ColorRGBA & color )
{
	glPushAttrib(GL_ENABLE_BIT | GL_LINE_BIT);
	glDisable(GL_LIGHTING);

	glLineWidth(m_fLineWidth);

	if (m_circleDisplayList == 0) {
		m_circleDisplayList = ShapeManager::GetCurrent()->GetDisplayListByShape( ShapeManager::Circle );
	}

	glColor4f( color.R(), color.G(), color.B(), color.A() );
	glPushMatrix();
	glTranslatef( m_circle.Center().X(), m_circle.Center().Y(), m_circle.Center().Z() );
	glMultMatrixf( &m_glRotationMatrix[0][0] );
	glScalef( m_circle.Radius(), m_circle.Radius(), m_circle.Radius() );
	glCallList( m_circleDisplayList );
	glPopMatrix();

	glPopAttrib();
}

bool Circle::Intersect( const Wml::Ray3f & ray, 
					    Wml::Vector3f & ptPosition, float & fDepth )
{
	// find closest point on circle
	Wml::Vector3f ptLineClosest, ptCircleClosest;

	// need to convert ray to line for Wml
	Wml::Line3f line;
	line.Origin() = ray.Origin();
	line.Direction() = ray.Direction();

	// find distance between line and circle
	float fDistanceSqr = Wml::SqrDistance(line, m_circle, &ptLineClosest, &ptCircleClosest);

	if (fDistanceSqr < Circle::s_IntersectionTolerance*Circle::s_IntersectionTolerance) {
		float fRayOrigin = ray.Direction().Dot(ray.Origin());
		float fRayPoint = ray.Direction().Dot(ptLineClosest);

		if (fRayPoint > fRayOrigin) {
			ptPosition = ptLineClosest;
			fDepth = fRayPoint;
			return true;
		}
	}
	return false;
}


//! initialize Wml circle from center point, normal, and radius
void Circle::InitializeCircle( const Wml::Vector3f & ptCenter,
							   const Wml::Vector3f & vNormal, float fRadius )
{
	// compute cosine of angle between vectors
	float axisDot = vNormal.Dot( Wml::Vector3f::UNIT_Z );

	// compute rotation axis
	bool bRotate = true;
	Wml::Vector3f axisCross( Wml::Vector3f::UNIT_Z.Cross( vNormal ) );

	// apply rotation if necessary
	Wml::Matrix3f matrix;
	if (axisCross.SquaredLength() > Wml::Mathf::EPSILON) {
		
		// compute normalized axis and angle, then create rotation around axis
		axisCross.Normalize();
		float fAngle = Wml::Mathf::ACos( axisDot / vNormal.Length() );
		matrix.FromAxisAngle( axisCross, fAngle );

	} else if (axisDot < 0.0) {
		matrix.FromAxisAngle( Wml::Vector3f::UNIT_X, 180.0f * Wml::Mathf::DEG_TO_RAD );
	} else {
		matrix = Wml::Matrix3f::IDENTITY;
	}

	// compute other axis for circle
	Wml::Vector3f axisU( matrix * Wml::Vector3f::UNIT_X );
	Wml::Vector3f axisV( matrix * Wml::Vector3f::UNIT_Y );

	// set up circle
	m_circle.Center() = ptCenter;
	m_circle.N() = vNormal;
	m_circle.U() = axisU;
	m_circle.V() = axisV;
	m_circle.Radius() = fRadius;

	// cache OpenGL rotation matrix 
	memset(m_glRotationMatrix, 0, sizeof(float)*16);
	for (int r = 0; r < 3; ++r) {
		m_glRotationMatrix[0][r] = matrix.GetRow(r)[0];
		m_glRotationMatrix[1][r] = matrix.GetRow(r)[1];
		m_glRotationMatrix[2][r] = matrix.GetRow(r)[2];
	}
	m_glRotationMatrix[3][3] = 1.0;
}






/*
 * Chevron
 */

const float Chevron::s_IntersectionTolerance = 20.0f;

void Chevron::Render( const Wml::ColorRGBA & color )
{
	glPushAttrib(GL_ENABLE_BIT | GL_LINE_BIT);
	glDisable(GL_LIGHTING);

	glLineWidth(m_fLineWidth);

	if (m_chevronDisplayList == 0) {
		m_chevronDisplayList = ShapeManager::GetCurrent()->GetDisplayListByShape( ShapeManager::Chevron );
	}

	glColor4f( color.R(), color.G(), color.B(), color.A() );
	glPushMatrix();
	glTranslatef( m_circle.Center().X(), m_circle.Center().Y(), m_circle.Center().Z() );
	glMultMatrixf( &m_glRotationMatrix[0][0] );
	glScalef( m_circle.Radius(), m_circle.Radius(), m_circle.Radius() );
	glRotatef( m_fAngle * Wml::Mathf::RAD_TO_DEG, 0, 0, 1.0f );
	glCallList( m_chevronDisplayList );
	glPopMatrix();

	glPopAttrib();
}

bool Chevron::Intersect( const Wml::Ray3f & ray, 
					     Wml::Vector3f & ptPosition, float & fDepth )
{
	//// find closest point on circle
	//Wml::Vector3f ptLineClosest, ptCircleClosest;

	//// need to convert ray to line for Wml
	//Wml::Line3f line;
	//line.Origin() = ray.Origin();
	//line.Direction() = ray.Direction();

	//// find distance between line and circle
	//float fDistanceSqr = Wml::SqrDistance(line, m_circle, &ptLineClosest, &ptCircleClosest);

	//if (fDistanceSqr < Circle::s_IntersectionTolerance*Circle::s_IntersectionTolerance) {
	//	float fRayOrigin = ray.Direction().Dot(ray.Origin());
	//	float fRayPoint = ray.Direction().Dot(ptLineClosest);

	//	if (fRayPoint > fRayOrigin) {
	//		ptPosition = ptLineClosest;
	//		fDepth = fRayPoint;
	//		return true;
	//	}
	//}
	return false;
}


//! initialize Wml circle from center point, normal, and radius
void Chevron::InitializeChevron( const Wml::Vector3f & ptCenter,
							     const Wml::Vector3f & vNormal, 
								 float fAngle, float fScale )
{
	// [RMS: isn't this just that rotate-to-Z function that is in vector utils now ??]

	// compute cosine of angle between vectors
	float axisDot = vNormal.Dot( Wml::Vector3f::UNIT_Z );

	// compute rotation axis
	bool bRotate = true;
	Wml::Vector3f axisCross( Wml::Vector3f::UNIT_Z.Cross( vNormal ) );

	// apply rotation if necessary
	Wml::Matrix3f matrix;
	if (axisCross.SquaredLength() > Wml::Mathf::EPSILON) {
		
		// compute normalized axis and angle, then create rotation around axis
		axisCross.Normalize();
		float fAngle = Wml::Mathf::ACos( axisDot / vNormal.Length() );
		matrix.FromAxisAngle( axisCross, fAngle );

	} else if (axisDot < 0.0) {
		matrix.FromAxisAngle( Wml::Vector3f::UNIT_X, 180.0f * Wml::Mathf::DEG_TO_RAD );
	} else {
		matrix = Wml::Matrix3f::IDENTITY;
	}

	// compute other axis for circle
	Wml::Vector3f axisU( matrix * Wml::Vector3f::UNIT_X );
	Wml::Vector3f axisV( matrix * Wml::Vector3f::UNIT_Y );

	// save plane information in circle class (it is just handy...)
	m_circle.Center() = ptCenter;
	m_circle.N() = vNormal;
	m_circle.U() = axisU;
	m_circle.V() = axisV;
	m_circle.Radius() = fScale;
	m_fAngle = fAngle;

	// cache OpenGL rotation matrix 
	memset(m_glRotationMatrix, 0, sizeof(float)*16);
	for (int r = 0; r < 3; ++r) {
		m_glRotationMatrix[0][r] = matrix.GetRow(r)[0];
		m_glRotationMatrix[1][r] = matrix.GetRow(r)[1];
		m_glRotationMatrix[2][r] = matrix.GetRow(r)[2];
	}
	m_glRotationMatrix[3][3] = 1.0;
}





/*
 * Cone
 */


unsigned int Cone::s_coneResolution = 10;	// hardcoded
unsigned int Cone::s_coneDisplayList = 0;
float Cone::s_fConeWidth = 1.0;	// hardcoded
float Cone::s_fConeHeight = 1.0;

void Cone::Render()
{
	glPushMatrix();

	Wml::Vector3f translate(BasePoint());
	glTranslatef( translate.X(), translate.Y(), translate.Z() );

	// compute cosine of angle between vectors
	float axisDot = m_cone.Axis().Dot( Wml::Vector3f::UNIT_Z );

	// compute rotation axis
	bool bRotate = true;
	Wml::Vector3f axisCross( Wml::Vector3f::UNIT_Z.Cross( m_cone.Axis() ) );

	// apply rotation if necessary
	if (axisCross.SquaredLength() > Wml::Mathf::EPSILON) {
		
		// compute normalized axis and angle, then create rotation around axis
		axisCross.Normalize();
		float fAngle = Wml::Mathf::ACos( axisDot / m_cone.Axis().Length() );
		Wml::Matrix3f matrix;
		matrix.FromAxisAngle( axisCross, fAngle );

		// create OpenGL rotation matrix 
		float glMatrix[4][4];
		memset(glMatrix, 0, sizeof(float)*16);
		for (int r = 0; r < 3; ++r) {
			glMatrix[0][r] = matrix.GetRow(r)[0];
			glMatrix[1][r] = matrix.GetRow(r)[1];
			glMatrix[2][r] = matrix.GetRow(r)[2];
		}
		glMatrix[3][3] = 1.0;
		glMultMatrixf( (GLfloat *)glMatrix );

	} else if (axisDot < 0.0) {
		glRotatef(180.0, 0.0, 1.0, 0.0);
	} else {
		// no rotation
	}

	glCallList( GetConeDisplayList() );
	glPopMatrix();
}

unsigned int Cone::GetConeDisplayList()
{
	if ( s_coneDisplayList == 0 ) {

		s_coneDisplayList = glGenLists(1);
		ASSERT(s_coneDisplayList != 0);

		glNewList( s_coneDisplayList, GL_COMPILE );

		GLUquadric * quadric = gluNewQuadric();
		gluCylinder( quadric, s_fConeWidth, 0.0, s_fConeHeight, s_coneResolution, s_coneResolution );
		gluDisk( quadric, 0.0f, 1.0f, s_coneResolution, s_coneResolution );
		gluDeleteQuadric( quadric );

		glEndList();

		ASSERT(glGetError() == GL_NO_ERROR);
	}
	return s_coneDisplayList;
}


