// Ryan Schmidt - rms@unknownroad.com
// Copyright (c) 2005-2010
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt


#include "StdAfx.h"
#include "Parameters.h"

#include <float.h>

using namespace rmsui;


IParameter * ParameterList::GetParameter( const char * pName ) const
{
	size_t nCount = m_vParams.size();
	CString stringName(pName);
	for ( unsigned int i = 0; i < nCount; ++i ) {
		if ( stringName == m_vParams[i]->GetName() )
			return m_vParams[i];
	}
	return NULL;
}

void ParameterList::Store( Wml::SerializerRecord * pRecord )
{
	size_t nCount = m_vParams.size();
	for ( unsigned int i = 0; i < nCount; ++i )
		m_vParams[i]->Store( pRecord );
}

void ParameterList::Restore( Wml::SerializerRecord * pRecord )
{
	size_t nCount = m_vParams.size();
	for ( unsigned int i = 0; i < nCount; ++i )
		m_vParams[i]->Restore( pRecord );
}




GenericParameter::GenericParameter( const CString & paramName, ParameterType paramType )
	: m_paramName(paramName), m_paramType(paramType)
{
}


void GenericParameter::GetValue( CString & value )
{
	switch (m_paramType) {
		case FloatParameter: value = CString("0.0"); break;
		case IntegerParameter: value = CString("0"); break;
		default: value = CString(""); break;
	}
}



/*
 * Generic Float Parameter
 */


GenericFloatParameter::GenericFloatParameter( const CString & paramName, float fInitialValue )
	: GenericParameter( paramName, FloatParameter ),
	  m_fParamValue(fInitialValue)
{
}

void GenericFloatParameter::GetValue( CString & value )
{
	value.Format("%f", GetValueFloat() );
}

void GenericFloatParameter::SetValue( const CString & string )
{
	char * pString = (char *)(LPCSTR)string;
	char * pStopString;
	int nLength = string.GetLength();
	float fValue = (float)strtod( pString, &pStopString );
	if (fValue == 0.0 && pStopString == pString) {
		// ignore! bad string!
	} else if ( Validate(fValue) == false ) {
		// ignore! bad string!
	} else {
		SendParameter(fValue);
		m_fParamValue = GetParameter();
	}
}

void GenericFloatParameter::SetValueFloat(float fValue)
{
	if ( Validate(fValue) == true ) {
		SendParameter(fValue);
		m_fParamValue = GetParameter();
	}
}

float GenericFloatParameter::GetMinValue()
{
	float fMin, fMax;
	GetMinMax(fMin,fMax);
	return fMin;
}

void GenericFloatParameter::GetMinMax( float & fMin, float & fMax )
{
	fMin = FLT_MIN;
	fMax = FLT_MAX;
}

bool GenericFloatParameter::Validate(float fValue)
{
	float fMin = 0, fMax = 0;
	GetMinMax(fMin, fMax);
	if (fValue < fMin || fValue > fMax)
		return false;
	return true;
}

void GenericFloatParameter::SynchronizeParameter()
{
	m_fParamValue = GetParameter();
}


/*
 * Generic Integer Parameter
 */



GenericIntegerParameter::GenericIntegerParameter( const CString & paramName, int nInitialValue )
	: GenericParameter( paramName, IntegerParameter ),
	  m_nParamValue(nInitialValue)
{
}

void GenericIntegerParameter::GetValue( CString & value )
{
	value.Format("%d", GetValueInteger() );
}

void GenericIntegerParameter::SetValue( const CString & string )
{
	char * pString = (char *)(LPCSTR)string;
	char * pStopString;
	int nLength = string.GetLength();
	int nValue = (int)strtol( pString, &pStopString, 10 );
	if (nValue == 0 && pStopString == pString) {
		// ignore! bad string!
	} else if ( Validate(nValue) == false ) {
		// ignore! bad string!
	} else {
		SendParameter(nValue);
		m_nParamValue = GetParameter();
	}
}

void GenericIntegerParameter::SetValueInteger(int nValue)
{
	if ( Validate(nValue) == true ) {
		SendParameter(nValue);
		m_nParamValue = GetParameter();
	}
}


void GenericIntegerParameter::GetMinMax( int & nMin, int & nMax )
{
	nMin = INT_MIN;
	nMax = INT_MAX;
}

bool GenericIntegerParameter::Validate(int nValue)
{
	int nMin = 0, nMax = 0;
	GetMinMax(nMin, nMax);
	if (nValue < nMin || nValue > nMax)
		return false;
	return true;
}

void GenericIntegerParameter::SynchronizeParameter()
{
	m_nParamValue = GetParameter();
}

