// Ryan Schmidt - rms@unknownroad.com
// Copyright (c) 2005-2010
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt


#ifndef __RMSUI_SCENE_OBJECT_H__
#define __RMSUI_SCENE_OBJECT_H__

#include <vector>

#include "Tool.h"
#include "Parameters.h"

// [RMS: these are locking macros for SceneObject. This is ugly, 
//   but since we're using boost threads, the only way to lock 
//   a scene object is to lock an internal mutex...
// [RMS] removed to get rid of boost dependency
//#define SCOPELOCK_SCENE_OBJECT(sceneObject)   boost::recursive_mutex::scoped_lock rmsui__solock( sceneObject->GetObjectMutex() )
#define SCOPELOCK_SCENE_OBJECT(sceneObject)   


namespace rmsui {

class SceneObject : public ITool3DTarget, public IParameterTarget
{
public:
	SceneObject(void) { m_pParent = NULL; m_eSelectionState = NotSelected; };
	virtual ~SceneObject(void) {};

	//! try to pick object. Returns NULL on failure, otherwise returns picked SceneObject (may be a child!)
	virtual SceneObject * Pick( const Wml::Ray3f & ray, float & fSqrDistance );

	//! render object
	virtual void Render(bool bSelected = false);


	/*
	 * state functions
	 */
	enum SelectionState {
		Selected,
		NotSelected
	};

	//! change selection state
	virtual void SetSelectionState( SelectionState eNewState );
	SelectionState GetSelectionState();

	/*
	 * ITool3DTarget functions
	 */

	//! get the frame for this target
	virtual Wml::Frame GetFrame();

	//! translate the frame
	virtual void Translate( const Wml::Vector3f & vTranslate, bool bRelative = true );

	//! rotate the frame
	virtual void Rotate( const Wml::Matrix3f & rotation );

	//! test if target supports a tool
	virtual bool SupportsTool( const std::string & sToolName );

	/*
	 * functions for dealing with children
	 */ 

	//! check if object has children
	bool HasChildren();

	//! add a child
	virtual void AddChild(SceneObject * pChild, int nIndex = -1);

	//! remove a child
	virtual void RemoveChild(SceneObject * pChild);

	//! remove all children
	virtual void RemoveAllChildren();

	//! get children
	const std::vector<SceneObject *> & GetChildren();

	/*
	 * functions for dealing with parent
	 */
	virtual void SetParent( SceneObject * pParent );

	virtual SceneObject * GetParent();



	/*
	 * functions for dealing with tool-children
	 */

	//! check if object has tool children
	bool HasToolChildren();

	//! add a tool child
	virtual void AddToolChild( SceneObject * pToolChild );

	//! remove a tool child
	virtual void RemoveToolChild( SceneObject * pToolChild );

	//! remove all tool children
	virtual void RemoveAllToolChildren();

	//! get tool children
	const std::vector<SceneObject *> & GetToolChildren();


	/*
	 * IParameterTarget functions  (interface and related)
	 */

	virtual const ParameterList & GetParameterSet();


	/*
	 * locking access
	 */
	//boost::recursive_mutex & GetObjectMutex();

protected:
	std::vector<SceneObject *> m_vChildren;
	std::vector<SceneObject *> m_vToolChildren;
	SceneObject * m_pParent;

	SelectionState m_eSelectionState;

	ParameterList m_parameters;

	inline virtual void AddParameter( IParameter * pParameters );

	//boost::recursive_mutex m_lockMutex;
};


// dummy implementations


inline SceneObject * SceneObject::Pick( const Wml::Ray3f & ray, float & fSqrDistance )
{
	// pick children

	// [RMS: disabled because scene object manager tries to pick all managed objects anyway!]
	//// [RMS TODO: should we return first hit child, or best-hit child ??]
	//size_t nChildren = m_vChildren.size();
	//for (unsigned int i = 0; i < nChildren; ++i) {
	//	float fHitDist;
	//	SceneObject * pHit = m_vChildren[i]->Pick(ray, fHitDist);
	//	if (pHit != NULL) {
	//		fSqrDistance = fHitDist;
	//		return pHit;
	//	}
	//}

	// pick tool children  
	if ( m_eSelectionState == Selected ) {
		size_t nToolChildren = m_vToolChildren.size();
		for (unsigned int i = 0; i < nToolChildren; ++i) {
			float fHitDist;
			SceneObject * pHit = m_vToolChildren[i]->Pick(ray, fHitDist);
			if (pHit != NULL) {
				fSqrDistance = fHitDist;
				return pHit;
			}
		}
	}

	return NULL;
}

inline void SceneObject::Render(bool bSelected)
{
	// [RMS: disabled because scene object manager renders all children anyway!]
	//// render children
	//size_t nChildren = m_vChildren.size();
	//for (unsigned int i = 0; i < nChildren; ++i)
	//	m_vChildren[i]->Render();

	// render tool children
	if ( m_eSelectionState == Selected ) {
		size_t nToolChildren = m_vToolChildren.size();
		for ( unsigned int i = 0; i < nToolChildren; ++i)
			m_vToolChildren[i]->Render();
	}
}

inline bool SceneObject::HasChildren()
{
	return ! m_vChildren.empty();
}

inline void SceneObject::AddChild(SceneObject * pChild, int nIndex)
{
	unsigned int nChildren = (unsigned int)m_vChildren.size();
	if ( nIndex == -1 || nIndex > (int)nChildren ) {
		m_vChildren.push_back(pChild);
	} else {
		m_vChildren.resize( nChildren + 1 );
		for ( unsigned int i = nChildren; i > (unsigned int)nIndex; --i )
			m_vChildren[i] = m_vChildren[i-1];
		m_vChildren[nIndex] = pChild;
	}

//	m_vChildren.push_back(pChild);
	pChild->SetParent(this);
}

inline void SceneObject::RemoveChild(SceneObject * pChild)
{
	size_t nSize = m_vChildren.size();
	for (unsigned int i = 0; i < nSize; ++i) {
		if (m_vChildren[i] == pChild) {
			for (unsigned int j = i; j < nSize; ++j)
				m_vChildren[j] = m_vChildren[j+1];
			m_vChildren.pop_back();
			i = (unsigned int)nSize;
		}
	}
}

inline void SceneObject::RemoveAllChildren()
{
	m_vChildren.clear();
}

inline const std::vector<SceneObject *> & SceneObject::GetChildren()
{
	return m_vChildren;
}


inline Wml::Frame SceneObject::GetFrame()
{
	return Wml::Frame();
}

inline void SceneObject::Translate( const Wml::Vector3f & vTranslate, bool bRelative )
{
}

inline void SceneObject::Rotate( const Wml::Matrix3f & rotation )
{
}

inline bool SceneObject::SupportsTool( const std::string & sToolName )
{
	return false;
}


inline  const ParameterList & SceneObject::GetParameterSet()
{
	return m_parameters;
}

inline void SceneObject::AddParameter( IParameter * pParameter )
{
	m_parameters.AddParameter( pParameter );
}



inline void SceneObject::SetParent( SceneObject * pParent )
{
	m_pParent = pParent;
}

inline SceneObject * SceneObject::GetParent()
{
	return m_pParent;
}




inline bool SceneObject::HasToolChildren()
{
	return ! m_vToolChildren.empty();
}

inline void SceneObject::AddToolChild(SceneObject * pChild)
{
	m_vToolChildren.push_back(pChild);
	pChild->SetParent(this);
}

inline void SceneObject::RemoveToolChild(SceneObject * pChild)
{
	size_t nSize = m_vToolChildren.size();
	for (unsigned int i = 0; i < nSize; ++i) {
		if (m_vToolChildren[i] == pChild) {
			for (unsigned int j = i; j < nSize; ++j)
				m_vToolChildren[j] = m_vChildren[j+1];
			m_vToolChildren.pop_back();
			i = (unsigned int)nSize;
		}
	}
}

inline void SceneObject::RemoveAllToolChildren()
{
	m_vToolChildren.clear();
}

inline const std::vector<SceneObject *> & SceneObject::GetToolChildren()
{
	return m_vToolChildren;
}


inline void SceneObject::SetSelectionState( SceneObject::SelectionState eNewState )
{
	m_eSelectionState = eNewState;
}

inline SceneObject::SelectionState SceneObject::GetSelectionState()
{
	return m_eSelectionState;
}

//inline boost::recursive_mutex & SceneObject::GetObjectMutex()
//{
//	return m_lockMutex;
//}

}  // end namespace rmsui


#endif  // __RMSUI_SCENE_OBJECT_H__