// Ryan Schmidt - rms@unknownroad.com
// Copyright (c) 2005-2010
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt


#include "StdAfx.h"
#include ".\sceneobjectmanager.h"

#include <float.h>

using namespace rmsui;


SceneObjectManager::SceneObjectManager(void)
{
}

SceneObjectManager::~SceneObjectManager(void)
{
}

bool SceneObjectManager::TestForObjectHit( const Wml::Ray3f & ray )
{
	// try to pick. Needs to happen before DeSelectAll() so that selected
	// objects still have proper state set (for tool children)
	SceneObject * pToolChildSelection = NULL;
	SceneEntity * pEntity = FindClosestHitEntity( ray, &pToolChildSelection );

	return pEntity != NULL;
}

bool SceneObjectManager::SelectObject( const Wml::Ray3f & ray, bool bAddToCurrent )
{
	// try to pick. Needs to happen before DeSelectAll() so that selected
	// objects still have proper state set (for tool children)
	SceneObject * pToolChildSelection = NULL;
	SceneEntity * pEntity = FindClosestHitEntity( ray, &pToolChildSelection );

	// clear selection
	if (!bAddToCurrent)
		DeSelectAll();

	if (pEntity) {
		pEntity->m_bSelected = true;
		pEntity->m_pObject->SetSelectionState( SceneObject::Selected );

		if ( pToolChildSelection ) {
			pEntity->m_bToolChildSelected = true;
			pEntity->m_pToolChildSelection = pToolChildSelection;
		}

		AddSelectedEntity( pEntity );
		return true;
	}
	return false;
}

bool SceneObjectManager::DeSelectObject( const Wml::Ray3f & ray )
{
	SceneObject * pToolChildSelection = NULL;
	SceneEntity * pEntity = FindClosestHitEntity( ray, &pToolChildSelection );

	if (pEntity) {
		pEntity->m_bSelected = false;
		pEntity->m_bToolChildSelected = false;
		pEntity->m_pToolChildSelection = NULL;
		pEntity->m_pObject->SetSelectionState( SceneObject::NotSelected );
		RemoveSelectedEntity( pEntity );
		return true;
	}
	return false;
}

bool SceneObjectManager::SelectObject( SceneObject * pObject )
{
	SceneEntity findEntity = {pObject,0};
	EntityList::iterator itr( m_entities.find(findEntity) );
	if (itr != m_entities.end()) {
		(*itr).m_bSelected = true;
		(*itr).m_pObject->SetSelectionState( SceneObject::Selected );
		AddSelectedEntity( &(*itr) );
		return true;
	}
	return false;
}


void SceneObjectManager::SelectAll()
{
	EntityList::iterator cur(m_entities.begin()), end(m_entities.end());
	while (cur != end) {
		(*cur).m_bSelected = true;
		(*cur).m_pObject->SetSelectionState( SceneObject::Selected );
		++cur;
	}
}

void SceneObjectManager::DeSelectAll()
{
	EntityList::iterator cur(m_entities.begin()), end(m_entities.end());
	while (cur != end) {
		(*cur).m_bSelected = false;
		(*cur).m_bToolChildSelected = false;
		(*cur).m_pToolChildSelection = NULL;
		(*cur).m_pObject->SetSelectionState( SceneObject::NotSelected );
		++cur;
	}

	ClearSelectedEntities();
}

void SceneObjectManager::Render()
{
	EntityList::iterator cur(m_entities.begin()), end(m_entities.end());
	while (cur != end) {
		(*cur).m_pObject->Render( (*cur).m_bSelected );
		++cur;
	}
}

void SceneObjectManager::AddObject( SceneObject * pObject, bool bSelected)
{
	SceneEntity entity;
	entity.m_pObject = pObject;
	entity.m_bSelected = bSelected;
	entity.m_bToolChildSelected = false;
	entity.m_pToolChildSelection = NULL;
	pObject->SetSelectionState( (bSelected) ? SceneObject::Selected : SceneObject::NotSelected );
	m_entities.insert(entity);

	if (bSelected)
		AddSelectedEntity(&entity);
}


void SceneObjectManager::RemoveObject( SceneObject * pObject )
{
	EntityList::iterator cur(m_entities.begin()), end(m_entities.end());
	while (cur != end) {
		if ( (*cur).m_pObject == pObject) {
			ASSERT( (*cur).m_bSelected == false);   // TODO: fix this...
			m_entities.erase( cur );
			return;
		}
		++cur;
	}
}


void SceneObjectManager::RemoveObjectAndChildren( SceneObject * pObject )
{
	// iterate through children
	const std::vector<SceneObject *> & vChildren = pObject->GetChildren();
	size_t nCount = vChildren.size();
	for ( unsigned int i = 0; i < nCount; ++i )
		RemoveObjectAndChildren( vChildren[i] );

	RemoveObject(pObject);
}


void SceneObjectManager::UpdateToolManager( Tool3DManager * pToolManager )
{
	pToolManager->BeginAddingToolTargets();
	EntityList::iterator cur(m_entities.begin()), end(m_entities.end());
	while (cur != end) {
		if ( (*cur).m_bSelected ) {

			if ( (*cur).m_bToolChildSelected )
				pToolManager->AddToolTarget( (*cur).m_pToolChildSelection );
			else
				pToolManager->AddToolTarget( (*cur).m_pObject );
		}
		++cur;
	}
	pToolManager->EndAddingToolTargets();
}


bool SceneObjectManager::GetSceneObjects( std::vector<SceneObject *> & vObjects )
{
	size_t vecSize = vObjects.size();
	size_t numEntities = m_entities.size();
	if (vecSize < numEntities)
		return false;

	unsigned int i = 0;
	EntityList::iterator cur(m_entities.begin()), end(m_entities.end());
	while (cur != end) {
		vObjects[i++] = (*(cur++)).m_pObject;
	}
	return true;
}



SceneObjectManager::SceneEntity * SceneObjectManager::FindClosestHitEntity( const Wml::Ray3f & ray, 
																		    SceneObject ** pToolChildSelection )
{
	EntityList::iterator cur(m_entities.begin()), end(m_entities.end());

	float fMinSqrDistance = FLT_MAX;
	SceneEntity * pHitEntity = NULL;
	SceneObject * pHitObject = NULL;

	while (cur != end) {
		SceneObject * pObject = (*cur).m_pObject;

		float fSqrDistance;
		SceneObject * pHit = pObject->Pick(ray, fSqrDistance);

		if (pHit) {

			if (fSqrDistance < fMinSqrDistance) {
				fSqrDistance = fMinSqrDistance;
				pHitEntity = &(*cur);
				pHitObject = pHit;
			}
		}

		++cur;
	}

	// check if tool child was selected
	if ( pHitEntity && pHitObject != pHitEntity->m_pObject ) {
		*pToolChildSelection = pHitObject;
	} else {
		*pToolChildSelection = NULL;
	}

	return pHitEntity;
}

bool SceneObjectManager::AddSelectedEntity( SceneEntity * pEntity )
{
	if ( m_selectedEntities.find(*pEntity) == m_selectedEntities.end()) {
		m_selectedEntities.insert(*pEntity);
		m_vecSelectedEntities.push_back(pEntity->m_pObject);
		return true;
	} else
		return false;
}

bool SceneObjectManager::RemoveSelectedEntity( SceneEntity * pEntity )
{
	if ( m_selectedEntities.find(*pEntity) != m_selectedEntities.end()) {

		pEntity->m_bToolChildSelected = false;
		pEntity->m_pToolChildSelection = NULL;

		m_selectedEntities.erase(*pEntity);
		m_vecSelectedEntities.clear();
		
		EntityList::iterator cur(m_selectedEntities.begin()), end(m_selectedEntities.end());
		while (cur != end)
			m_vecSelectedEntities.push_back( (*cur++).m_pObject );
		return true;
	} else
		return false;
}

void SceneObjectManager::ClearSelectedEntities()
{
	m_selectedEntities.clear();
	m_vecSelectedEntities.clear();
}	
