// Ryan Schmidt - rms@unknownroad.com
// Copyright (c) 2005-2010
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt


#ifndef __RMSUI_TOOL_MANAGER_H__
#define __RMSUI_TOOL_MANAGER_H__

#include <vector>
#include <ExtendedWmlCamera.h>

#include "InputEventStream.h"
#include "Tool.h"

#include <set>

namespace rmsui
{



class IToolActionEventListener
{
public:
	virtual ~IToolActionEventListener() {}

	virtual void OnToolMoveEvent() = 0;
};


/**
 *
 */
class Tool3DManager 
	: public IInputEventListener,
	  public ITool3DListener
{
public:
	Tool3DManager()
		: m_bInputCaptured(false),
		  m_bSelected(false),
		  m_bToolEnabled(false),
		  m_pTool(NULL)
	{
	}

	~Tool3DManager()
	{
		RemoveAllTools();
	}

/*
 * IInputEventListener interface
 */
	virtual EventState ProcessInputEvent( const rmsui::InputEventData & event );

/*
 * ITool3DListener interface
 */
	virtual void OnToolActionApplied( ITool3D * pTool );

/*
 * event support
 */
	void AddListener( IToolActionEventListener * pListener ) { m_vListeners.insert(pListener); }
	void RemoveListener( IToolActionEventListener * pListener ) { m_vListeners.erase(pListener); }


	//! set current camera
	void SetCamera( Wml::ExtendedWmlCamera * pCamera )
		{ m_pCamera = pCamera; }

	//! get current camera
	Wml::ExtendedWmlCamera * GetCamera() 
		{ return m_pCamera; }

	//! set current tool target set
	void SetToolTarget( ITool3DTarget * pTarget )
		{ m_vTargetGroup.Clear(); m_vTargetGroup.AddTarget(pTarget); UpdateTool(); }

	//! set current tool target set
	void SetToolTargets( const std::vector<ITool3DTarget *> & vToolTargets )
		{m_vTargetGroup.SetToolTargets(vToolTargets); UpdateTool(); }

	//! begin multi-call iterative tool target adding
	void BeginAddingToolTargets()
		{ m_vTargetGroup.Clear(); }

	//! continue iterative tool target adding
	void AddToolTarget( ITool3DTarget * pTarget )
		{ m_vTargetGroup.AddTarget(pTarget); }

	//! finish multi-call iterative tool target adding
	void EndAddingToolTargets()
		{ UpdateTool(); }

	void ClearToolTargets()
		{ m_vTargetGroup.Clear(); UpdateTool(); }



/*
 * tool control
 */

	//! add a tool
	void RegisterTool( ITool3D * pTool );

	//! set current tool
	void SelectTool( const std::string & sToolName );

	//! select next applicable tool
	void SelectNextValidTool();

	//! set current tool state - when tool is enabled it is drawn and can be selected
	void EnableCurrentTool( bool bEnable );


	// rendering support

	void Render( );

	// interaction support

	bool PickTool( int screenX, int screenY );

	void MoveTool( int screenX, int screenY );

	void SetSelected( bool bSelected )
		{ m_bSelected = bSelected; }

	bool GetSelected()
		{ return m_bSelected; }



protected:

	//! current camera
	Wml::ExtendedWmlCamera * m_pCamera;

	//! flag set if we current have captured input
	bool m_bInputCaptured;

	//! selection flag
	bool m_bSelected;

	//! when enabled, current tool is drawn and can be selected / interacted
	bool m_bToolEnabled;

	//! current tool
	ITool3D * m_pTool;

	//! list of available tools
	ToolSet m_toolSet;

	//! multi-selection target set
	ToolTargetGroup m_vTargetGroup;

	void UpdateTool();

	void RemoveAllTools() { m_toolSet.ClearTools(); }

/*
 * event support
 */
	//! listeners
	std::set<IToolActionEventListener *> m_vListeners;
	void NotifyToolMoveEvent();

/*
 * rendering support
 */
	//! current view state - opengl matrices, etc
	ViewState m_viewState;

	//! info about selected target (RMS TODO improve this? remove? hrm...)
	SelectedTargetInfo m_targetInfo;


	//! rendering support function - set new frustum so tools are fixed-size
	void SetupFixedSizeView( );	

	/*! compute translation to tool that places it's origin at the correct position
	    in the fixed-size perspective space */
	void ComputeFixedSizeTranslation( const ITool3D * pTool,
									  Wml::Vector3f & nearPlanePoint, 
									  Wml::Vector3f & directionIntoScene );

	//! generate a ray using the given modelview and projection matrices
	void GenerateEyeRay( int nScreenX, int nScreenY, 
						 double dModelViewMatrix[16], double dProjectionMatrix[16], 
						 Wml::Ray3f & ray );

	//! transform a fixed-view space ray into the local space of a tool
	void TransformRayIntoToolSpace( const ITool3D * pTool, const Wml::Ray3f & initialRay,
									const Wml::Matrix3f & matViewOrientation,
									Wml::Ray3f & destRay );

};



}	// namespace rmsui



#endif  /// __RMSUI_TOOL_MANAGER_H__