// Ryan Schmidt - rms@unknownroad.com
// Copyright (c) 2004-2010
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt


#include "StdAfx.h"

#include "WmlExtSmoothCameraInterpolator.h"
#include "WmlExtVectorUtil.h"

using namespace Wml;


SmoothCameraInterpolator::SmoothCameraInterpolator()
{
	m_pCamera = NULL;
}

SmoothCameraInterpolator::SmoothCameraInterpolator( ExtendedWmlCamera * pCamera, 
												    const Wml::Vector3f & newLocation, 
													const Wml::Vector3f & newTarget,
													unsigned int nInterpSteps )
{
	m_nInterpSteps = nInterpSteps;
	m_nStepsElapsed = 0;

	m_pCamera = pCamera;
	m_vInitialUp = m_pCamera->GetUp();
	m_vInitialLeft = m_pCamera->GetLeft();
	m_vInitialDirection = m_pCamera->GetDirection();
	m_vInitialTarget = m_pCamera->GetTarget();

	m_fInitialDistance = (m_pCamera->GetTarget() - m_pCamera->GetLocation()).Length();

	m_vFinalTarget = newTarget;
	Wml::Vector3f vNewLookAt = newTarget - newLocation;
	m_fFinalDistance = vNewLookAt.Length();

	// compute new lookat vector
	Wml::Vector3f vOldLookAt = m_pCamera->GetDirection();
	float fOldDist = (m_pCamera->GetTarget() - m_pCamera->GetLocation()).Length();
	float fNewDist = vNewLookAt.Length();

	// compute perpendicular vector between current camera lookat and newLookAt
	m_vPerp = vNewLookAt.Cross(m_vInitialDirection);
		//m_vInitialDirection.Cross( vNewLookAt );
	m_vPerp.Normalize();

	// compute matrix that rotates perpedicular vector into z axis
	Wml::Matrix3f matIntoZ;
	Wml::ComputeAlignZAxisMatrix( m_vPerp, matIntoZ, true );

	// rotate lookat vectors into XY plane
	vOldLookAt = matIntoZ * vOldLookAt;
	vNewLookAt = matIntoZ * vNewLookAt;

	// compute angles around axis
	m_fStartAngle = std::atan2( vOldLookAt.Y(), vOldLookAt.X() );
	float fNewAngle = std::atan2( vNewLookAt.Y(), vNewLookAt.X() );

	// make quaterions
	m_qStart = Wml::Quaternionf(Wml::Vector3f::UNIT_Z, m_fStartAngle );
	m_qEnd = Wml::Quaternionf(Wml::Vector3f::UNIT_Z, fNewAngle );
}


SmoothCameraInterpolator::~SmoothCameraInterpolator()
{
}

bool SmoothCameraInterpolator::RunSlerpStep()
{
	if (m_pCamera == NULL) return false;

	m_nStepsElapsed++;
	float fCurParam = (float)m_nStepsElapsed / (float)m_nInterpSteps;
		
	Wml::Quaternionf curQuat( 
		Wml::Quaternionf::Slerp( fCurParam, m_qStart, m_qEnd ) );

	Wml::Vector3f vCurRotAxis;
	float fCurRotAngle;
	curQuat.ToAxisAngle( vCurRotAxis, fCurRotAngle );

	fCurRotAngle -= m_fStartAngle;
	Wml::Matrix3f axisRot;
	axisRot.FromAxisAngle( m_vPerp, fCurRotAngle );

	// compute interpolated target point
	Wml::Vector3f vTarget( m_vInitialTarget * (1.0f-fCurParam) + m_vFinalTarget * fCurParam );

	// compute interpolated distance
	float fDistance = m_fInitialDistance * (1.0f-fCurParam) + m_fFinalDistance * fCurParam;


	Wml::Vector3f vLeft( axisRot * m_vInitialLeft );
	Wml::Vector3f vUp( axisRot * m_vInitialUp );
	Wml::Vector3f vDirection( axisRot * m_vInitialDirection );

	m_pCamera->SetTargetFrame( 
		vTarget - (vDirection * fDistance), 
		vLeft, vUp, vTarget );

	return m_nStepsElapsed < m_nInterpSteps;
}